;; Generated package description from hermes.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "hermes" "0.3.1.0.20260815.0" "Emacs frontend for Hermes Agent" '((emacs "29.1") (keymap-popup "0.3.1") (websocket "1.15") (markdown-mode "2.6")) :commit "93def9ce6793e8bb7f4e47cec0d9fd04ec302100" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("tools" "convenience") :url "https://git.thanosapollo.org/emacs-hermes")
